import torch

class CTCLabelConverter(object):
    """ Convert between text-label and text-index """

    __device__ = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    def __init__(self, character):
        # character (str): set of the possible characters.
        dict_character = list(character)

        self.dict = {}
        for i, char in enumerate(dict_character):
            # NOTE: 0 is reserved for 'CTCblank' token required by CTCLoss
            self.dict[char] = i + 1

        self.character = ['[CTCblank]'] + dict_character  # dummy '[CTCblank]' token for CTCLoss (index 0)

    def encode(self, text, batch_max_length=25):
        """convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]
            batch_max_length: max length of text label in the batch. 25 by default

        output:
            text: text index for CTCLoss. [batch_size, batch_max_length]
            length: length of each text. [batch_size]
        """
        length = [len(s) for s in text]

        # The index used for padding (=0) would not affect the CTC loss calculation.
        batch_text = torch.LongTensor(len(text), batch_max_length).fill_(0)
        for i, t in enumerate(text):
            text = list(t)
            text = [self.dict[char] for char in text]
            batch_text[i][:len(text)] = torch.LongTensor(text)
        return (batch_text.to(CTCLabelConverter.__device__), torch.IntTensor(length).to(CTCLabelConverter.__device__))

    def decode(self, text_index, length):
        """ convert text-index into text-label. """
        texts = []
        for index, l in enumerate(length):
            t = text_index[index, :]

            char_list = []
            for i in range(l):
                if t[i] != 0 and (not (i > 0 and t[i - 1] == t[i])):  # removing repeated characters and blank.
                    char_list.append(self.character[t[i]])
            text = ''.join(char_list)

            texts.append(text)
        return texts


class CTCLabelConverterForBaiduWarpctc(object):
    """ Convert between text-label and text-index for baidu warpctc """

    def __init__(self, character):
        # character (str): set of the possible characters.
        dict_character = list(character)

        self.dict = {}
        for i, char in enumerate(dict_character):
            # NOTE: 0 is reserved for 'CTCblank' token required by CTCLoss
            self.dict[char] = i + 1

        self.character = ['[CTCblank]'] + dict_character  # dummy '[CTCblank]' token for CTCLoss (index 0)

    def encode(self, text, batch_max_length=25):
        """convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]
        output:
            text: concatenated text index for CTCLoss.
                    [sum(text_lengths)] = [text_index_0 + text_index_1 + ... + text_index_(n - 1)]
            length: length of each text. [batch_size]
        """
        length = [len(s) for s in text]
        text = ''.join(text)
        text = [self.dict[char] for char in text]

        return (torch.IntTensor(text), torch.IntTensor(length))

    def decode(self, text_index, length):
        """ convert text-index into text-label. """
        texts = []
        index = 0
        for l in length:
            t = text_index[index:index + l]

            char_list = []
            for i in range(l):
                if t[i] != 0 and (not (i > 0 and t[i - 1] == t[i])):  # removing repeated characters and blank.
                    char_list.append(self.character[t[i]])
            text = ''.join(char_list)

            texts.append(text)
            index += l
        return texts


class AttnLabelConverter(object):
    """ Convert between text-label and text-index """

    def __init__(self, character):
        # character (str): set of the possible characters.
        # [GO] for the start token of the attention decoder. [s] for end-of-sentence token.
        list_token = ['[GO]', '[s]']  # ['[s]','[UNK]','[PAD]','[GO]']
        list_character = list(character)
        self.character = list_token + list_character

        self.dict = {}
        for i, char in enumerate(self.character):
            # print(i, char)
            self.dict[char] = i

    def encode(self, text, batch_max_length=25):
        """ convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]
            batch_max_length: max length of text label in the batch. 25 by default

        output:
            text : the input of attention decoder. [batch_size x (max_length+2)] +1 for [GO] token and +1 for [s] token.
                text[:, 0] is [GO] token and text is padded with [GO] token after [s] token.
            length : the length of output of attention decoder, which count [s] token also. [3, 7, ....] [batch_size]
        """
        length = [len(s) + 1 for s in text]  # +1 for [s] at end of sentence.
        # batch_max_length = max(length) # this is not allowed for multi-gpu setting
        batch_max_length += 1
        # additional +1 for [GO] at first step. batch_text is padded with [GO] token after [s] token.
        batch_text = torch.LongTensor(len(text), batch_max_length + 1).fill_(0)
        for i, t in enumerate(text):
            text = list(t)
            text.append('[s]')
            text = [self.dict[char] for char in text]
            batch_text[i][1:1 + len(text)] = torch.LongTensor(text)  # batch_text[:, 0] = [GO] token
        return (batch_text.to(device), torch.IntTensor(length).to(device))

    def decode(self, text_index, length):
        """ convert text-index into text-label. """
        texts = []
        for index, l in enumerate(length):
            text = ''.join([self.character[i] for i in text_index[index, :]])
            texts.append(text)
        return texts



__all__ = ["ImageUtils"]

import cv2
import math
import imutils
import numpy as np
from skimage import io


class ImageUtils:

    @staticmethod
    def align_image(image:np.array, template:np.array, maxFeatures=500, keepPercent=0.2, debug=False
                    ) -> np.array:
        """
        If a good receipt template can be found use it.

        Args:
            image (np.array): Original image
            template (np.array): Template to be warped to
            maxFeatures (int, optional): . Defaults to 500.
            keepPercent (float, optional): . Defaults to 0.2.
            debug (bool, optional): . Defaults to False.

        Returns:
            [np.array]: allgined image
        """
        # convert both the input image and template to grayscale
        imageGray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        templateGray = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)

        # use ORB to detect keypoints and extract (binary) local
        # invariant features
        orb = cv2.ORB_create(maxFeatures)
        (kpsA, descsA) = orb.detectAndCompute(imageGray, None)
        (kpsB, descsB) = orb.detectAndCompute(templateGray, None)

        # match the features
        method = cv2.DESCRIPTOR_MATCHER_BRUTEFORCE_HAMMING
        matcher = cv2.DescriptorMatcher_create(method)
        matches = matcher.match(descsA, descsB, None)

        # sort the matches by their distance (the smaller the distance,
        # the "more similar" the features are)
        matches = sorted(matches, key=lambda x:x.distance)
        # keep only the top matches
        keep = int(len(matches) * keepPercent)
        matches = matches[:keep]
        # check to see if we should visualize the matched keypoints
        if debug:
            matchedVis = cv2.drawMatches(image, kpsA, template, kpsB,
                matches, None)
            matchedVis = imutils.resize(matchedVis, width=1000)
            cv2.imshow("Matched Keypoints", matchedVis)
            cv2.waitKey(0)

        # allocate memory for the keypoints (x, y)-coordinates from the
        # top matches -- we'll use these coordinates to compute our
        # homography matrix
        ptsA = np.zeros((len(matches), 2), dtype="float")
        ptsB = np.zeros((len(matches), 2), dtype="float")
        # loop over the top matches
        for (i, m) in enumerate(matches):
            # indicate that the two keypoints in the respective images
            # map to each other
            ptsA[i] = kpsA[m.queryIdx].pt
            ptsB[i] = kpsB[m.trainIdx].pt

        # compute the homography matrix between the two sets of matched
        # points
        (H, mask) = cv2.findHomography(ptsA, ptsB, method=cv2.RANSAC)
        # use the homography matrix to align the images
        (h, w) = template.shape[:2]
        aligned = cv2.warpPerspective(image, H, (w, h))
        # return the aligned image
        return aligned


    @staticmethod
    def allign_bboxes(img_shape:int, bboxes:np.array, divide_imagey:int = 40, divide_imagex:int = 30):
        """
        Allign boxes with a left-to-right and top-to-bottom manner. Draws grids with respect to x and y axis and first

        Args:
            img_shape (int): The shape of the image (h, w)
            bboxes (np.array): The bounding boxes (x, y)
            divide_imagey (int, optional): How much lines will be drawn in y axis. Defaults to 40.
            divide_imagex (int, optional): How much lines will be drawn in x axis. Defaults to 30.

        Returns:
            [type]: [description]
        """
        top_lefts = bboxes[:, 0, :]
        bottom_lefts = bboxes[:, -1, :]
        top_rights = bboxes[:, 1, :]
        all_y_dims = [np.linalg.norm(top_left - bottom_left) for top_left, bottom_left in zip(top_lefts, bottom_lefts)]
        all_x_dims = [np.linalg.norm(top_left - top_right) for top_left, top_right in zip(top_lefts, top_rights)]
        mean_y, std_y = math.ceil(np.mean(all_y_dims)), math.ceil(np.std(all_y_dims))
        mean_x, std_x = math.ceil(np.mean(all_x_dims)), math.ceil(np.std(all_x_dims))

        sorted_bboxes = [x for _,x in sorted(zip(bboxes[:, 0, 1],bboxes))]

        low_y = 0
        alligned_bboxes = []

        for y in range(0, img_shape[0] + mean_y + std_y,  mean_y + std_y):
            y_alligned = [bbox for bbox in sorted_bboxes if bbox[0,1] >= low_y and bbox[0,1] <= y]
            for x in range(0, img_shape[1] + mean_x  + std_x, mean_x + std_x):
                if y_alligned:
                    x_alligned = sorted([bbox for bbox in y_alligned if bbox[0,0] >= low_x and bbox[0,0] <= x], key=lambda x: x[0,0])
                    if x_alligned:
                        alligned_bboxes.extend(sorted(x_alligned, key=lambda x: x[0,1]))
                low_x = x
            low_y = y
        return alligned_bboxes

    @staticmethod
    def loadImage(img_file):
        img = io.imread(img_file)           # RGB order
        if img.shape[0] == 2: img = img[0]
        if len(img.shape) == 2 : img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        if img.shape[2] == 4:   img = img[:,:,:3]
        img = np.array(img)

        return img

    @staticmethod
    def cvt2HeatmapImg(img):
        img = (np.clip(img, 0, 1) * 255).astype(np.uint8)
        img = cv2.applyColorMap(img, cv2.COLORMAP_JET)
        return img

    @staticmethod
    def crop_image(img, bboxes:np.array, divide_imagey:int = 40, divide_imagex:int = 30):
        """
        Crop image areas where bounding boxes exist.

        Args:
            img ([np.array]): BGR image
            bboxes (np.array): Bounding boxes where x, y
            divide_imagey (int, optional): How much lines will be drawn in y axis. Defaults to 40.
            divide_imagex (int, optional): How much lines will be drawn in x axis. Defaults to 30.
        """

        alligned_bboxes = ImageUtils.allign_bboxes(img.shape, bboxes, divide_imagey, divide_imagex)
        cropped_images = []
        for box in alligned_bboxes:
            poly = np.array(box).astype(np.int32)
            ## (1) Crop the bounding rect
            rect = cv2.boundingRect(poly)
            x,y,w,h = rect
            croped = img[y:y+h, x:x+w].copy()

            ## (2) make mask
            poly = poly - poly.min(axis=0)

            mask = np.zeros(croped.shape[:2], np.uint8)
            cv2.drawContours(mask, [poly], -1, (255, 255, 255), -1, cv2.LINE_AA)

            ## (3) do bit-op
            dst = cv2.bitwise_and(croped, croped, mask=mask)

            # ## (4) add the white background
            bg = np.ones_like(croped, np.uint8)*255
            cv2.bitwise_not(bg,bg, mask=mask)
            dst2 = bg+ dst

            cropped_images.append(dst2)
        return cropped_images



import os
import re
import logging
from typing import List

import torch
import numpy as np
from torch import nn
import torch.utils.data
import torch.nn.functional as F

from utils import FileHandler
from utils import ImageUtils
from utils import copyStateDict
from downloaders import download_model
from datasets import AlignCollate, RawDataset
from analyzer import Metrics, ImageManipulate
from models import Craft, RefineNet, Tesseract, ReNBAtt
from transforms import ImageTransform, PostTransform, AttnLabelConverter, CTCLabelConverter

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("receiptrecognizer.module")

# TODO: Add functionality for working as seperate models one can actually load recognition model
# without detection model but now first the class should be initialized by first loading the detection
# model then other models can be loaded.

class ReceiptRecognizer(nn.Module):

    __models_path__  = os.path.join(os.path.expanduser("~"), "receipt_models")

    __device__ = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    __gdrive_paths__ = {"craft_mlt_25k.pth": {
                                                "model": Craft,
                                                "model_id": "1Jk4eGD7crsqCCg9C9VjCLkMN3ze8kutZ"
                                            },
                        "craft_refiner_CTW1500.pth": {
                                                        "model": RefineNet,
                                                        "model_id": "1XSaFwBkOaFOdtk4Ane3DFyJGPRw6v5bO"
                                                    },
                        "TPS-ResNet-BiLSTM-Attn.pth": {
                                                        "model": ReNBAtt,
                                                        "model_id": "15WPsuPJDCzhp2SvYZLRj8mAlT3zmoAMW"
                                                    }
                        }


    __character__ = "0123456789abcdefghijklmnopqrstuvwxyz"

    __model_types__ = ["detection", "recognition"]

    __numclass__ = None

    def __init__(self, detector:torch.nn.Module, recognizer:torch.nn.Module = None):
        super().__init__()
        self.detector = detector
        self.refiner = None
        self.recognizer = None

    @classmethod
    def build(cls, model:torch.nn.Module):
        return cls(model)

    @classmethod
    def load(cls, path:str, model_name:str, model_type:str = "detection") -> torch.nn.Module:
        assert model_type in ReceiptRecognizer.__model_types__, ("Given model type is not in supported model types",
                                                                f"Supported model types: {ReceiptRecognizer.__model_types__}")
        if model_type == "detection":
            model = ReceiptRecognizer.__gdrive_paths__[model_name]["model"]()
        elif model_type == "recognition":
            model = ReceiptRecognizer.__gdrive_paths__[model_name]["model"](ReceiptRecognizer.__numclass__)
        model.load_state_dict(copyStateDict(torch.load(path, map_location = ReceiptRecognizer.__device__)))
        return model

    @classmethod
    def load_from_drive(cls, model_name:str, model_type:str = "detection") -> torch.nn.Module:
        model_path = download_model(ReceiptRecognizer.__models_path__, file_name = model_name,
                                    file_id = ReceiptRecognizer.__gdrive_paths__[model_name]["model_id"])
        logger.info(f"Model downloaded to {model_path}")
        return cls.load(model_path, model_name, model_type)

    @classmethod
    def load_pretrained_model(cls, model_name:str, model_path:str = None, local_path:str = None, model_type:str = "detection") -> torch.nn.Module:
        """
        Loads the pretrained model from either by downloading from google drive or finding the reserved model path or
        by taking a local path.

        Args:
            model_name (str): The basename of the model.
            model_path (str): The reserved path of the model
            local_path (str, optional): Models path in the local computer. Defaults to None.

        Returns:
            torch.nn.Module: PyTorch model.
        """
        if not model_path: model_path = os.path.join(ReceiptRecognizer.__models_path__, model_name)
        if not local_path and not os.path.exists(model_path):
            model = cls.load_from_drive(model_name)
        elif local_path:
            if not os.path.exists(local_path):
                logger.info(f"Given model path does not exists -> {local_path}\nDownloading from drive...")
                model = cls.load_from_drive(model_name)
            model = cls.load(local_path, model_name, model_type)
        else:
            model = cls.load(model_path, model_name, model_type)

        return model

    @classmethod
    def from_pretrained(cls, model_name:str, local_path:str = None):
        assert isinstance(model_name, str), "Model name is required as string"
        assert model_name in ReceiptRecognizer.__gdrive_paths__.keys(), ("Model name not found"
                                    f"Models we have {str(ReceiptRecognizer.__gdrive_paths__.keys())}")

        if not os.path.exists(ReceiptRecognizer.__models_path__):
            os.makedirs(ReceiptRecognizer.__models_path__, exist_ok = True)

        model_path = os.path.join(ReceiptRecognizer.__models_path__, model_name)

        return cls.build(cls.load_pretrained_model(model_name, model_path, local_path))


    @torch.no_grad()
    @torch.jit.unused
    def detection(self, image:np.ndarray, canvas_size:int = 1280, mag_ratio:float = 1.5, text_threshold:float = 0.7,
                link_threshold:float = 0.4, low_text:float = 0.4, poly:bool = False, refine:bool = True, **kwargs):

        img_resized, target_ratio, size_heatmap = ImageTransform.resize_aspect_ratio(image, canvas_size,
                                                        interpolation=cv2.INTER_LINEAR, mag_ratio=mag_ratio)
        ratio_h = ratio_w = 1 / target_ratio

        # preprocessing
        x = ImageTransform.normalizeMeanVariance(img_resized)
        x = torch.from_numpy(x).permute(2, 0, 1) # [h, w, c] to [c, h, w]
        x = x.unsqueeze(0) # [c, h, w] to [b, c, h, w]
        x.to(ReceiptRecognizer.__device__)

        # forward pass
        y, feature = self.detector(x)

        # make score and link map
        heatmap = y[0,:,:,0].cpu().data.numpy()
        score_link = y[0,:,:,1].cpu().data.numpy()

        # # refine link
        if refine:
            if not self.refiner:
                model_name = kwargs.get("refiner_name", "craft_refiner_CTW1500.pth")
                self.refiner = self.load_pretrained_model(model_name, os.path.join(ReceiptRecognizer.__models_path__, "craft_refiner_CTW1500.pth"))
                self.refiner.eval()
                y_refiner = self.refiner.forward(y, feature)
                score_link = y_refiner[0,:,:,0].cpu().data.numpy()
            else:
                y_refiner = self.refiner(y, feature)
                score_link = y_refiner[0,:,:,0].cpu().data.numpy()

        # Post-processing
        boxes, polys = PostTransform.getDetBoxes(heatmap, score_link, text_threshold, link_threshold, low_text, poly)

        # coordinate adjustment
        boxes = PostTransform.adjustResultCoordinates(boxes, ratio_w, ratio_h)
        polys = PostTransform.adjustResultCoordinates(polys, ratio_w, ratio_h)
        for k in range(len(polys)):
            if polys[k] is None: polys[k] = boxes[k]

        # render results (optional)
        render_img = heatmap.copy()
        render_img = np.hstack((render_img, score_link))
        ret_heatmap = ImageUtils.cvt2HeatmapImg(render_img)

        return boxes, polys, ret_heatmap

    @torch.no_grad()
    @torch.jit.unused
    def recognition(self, prediction_style:str = "Attn", imgH:int = 32, imgW:int = 100, pad:bool = False, rgb:bool = True,
                    image_folder:str = "../test_instances/results_detection", batch_size:int = 8, batch_max_length:int = 25):

        if not ReceiptRecognizer.__numclass__:
            if 'CTC' in prediction_style:
                self.converter = CTCLabelConverter(ReceiptRecognizer.__character__)
            else:
                self.converter = AttnLabelConverter(ReceiptRecognizer.__character__)

        if not self.recognizer:
            ReceiptRecognizer.__numclass__ = len(self.converter.character)
            self.recognizer = self.load_pretrained_model("TPS-ResNet-BiLSTM-Attn.pth", model_type = "recognition")
        self.recognizer.eval()

        # prepare data. two demo images from https://github.com/bgshih/crnn#run-demo
        AlignCollate_demo = AlignCollate(imgH = imgH, imgW = imgW, keep_ratio_with_pad = pad)
        demo_data = RawDataset(root=image_folder)  # use RawDataset
        demo_loader = torch.utils.data.DataLoader(
            demo_data, batch_size=batch_size,
            shuffle=False,
            collate_fn=AlignCollate_demo, pin_memory=True)

        result = []
        for image_tensors, image_path_list in demo_loader:
            batch_size = image_tensors.size(0)
            image = image_tensors.to(ReceiptRecognizer.__device__)
            # For max length prediction
            length_for_pred = torch.IntTensor([batch_max_length] * batch_size).to(ReceiptRecognizer.__device__)
            text_for_pred = torch.LongTensor(batch_size, batch_max_length + 1).fill_(0).to(ReceiptRecognizer.__device__)

            if 'CTC' in prediction_style:
                preds = self.recognizer(image, text_for_pred)

                # Select max probabilty (greedy decoding) then decode index to character
                preds_size = torch.IntTensor([preds.size(1)] * batch_size)
                _, preds_index = preds.max(2)
                # preds_index = preds_index.view(-1)
                preds_str = self.converter.decode(preds_index, preds_size)

            else:
                preds = self.recognizer(image, text_for_pred, is_train=False)

                # select max probabilty (greedy decoding) then decode index to character
                _, preds_index = preds.max(2)
                preds_str = self.converter.decode(preds_index, length_for_pred)

            preds_prob = F.softmax(preds, dim=2)
            preds_max_prob, _ = preds_prob.max(dim=2)
            for img_name, pred, pred_max_prob in zip(image_path_list, preds_str, preds_max_prob):
                if 'Attn' in prediction_style:
                    pred_EOS = pred.find('[s]')
                    pred = pred[:pred_EOS]  # prune after "end of sentence" token ([s])
                    pred_max_prob = pred_max_prob[:pred_EOS]

                # calculate confidence score (= multiply of pred_max_prob)
                confidence_score = pred_max_prob.cumprod(dim=0)[-1]
                result.append(pred)
        return result

    def allign_char_results(self, tesseract_results:List, model_results:List):
        final_result = []
        for tes, model in zip(tesseract_results, model_results):
            if tes == "\x0c":
                dollar_matches = list(re.finditer("s\d*", model))
                if len(dollar_matches) > 0:
                    new_model = ""
                    for dollar_match in dollar_matches:
                        if dollar_match.span()[1] - dollar_match.span()[0] > 2:
                            model = list(model)
                            model[dollar_match.span()[0]] = "$"
                            new_model += "".join(model[:dollar_match.span()[1]-2]) + "." + "".join(model[dollar_match.span()[1]-2:])
                        else:
                            new_model = model
                        final_result.append(new_model)
                else:
                    final_result.append(model)
            else:
                tes = re.sub(r"[\n\x0c]","",tes)
                final_result.append(tes)
        return final_result

if __name__ == "__main__":
    import cv2
    model = ReceiptRecognizer.from_pretrained("craft_mlt_25k.pth")
    model.detector.eval()
    test_path = "../test_instances/raw_images"
    results_path = "../test_instances/results_detection"
    # all_results = []
    # ground_truths = FileHandler.load_gt("/home/kemalaraz/Downloads/labels_my-project-name_2021-06-15-12-40-33.csv")
    # image_ious = {}
    # for k, image_path in enumerate(os.listdir(test_path)):
    #     image_path = os.path.join(test_path, image_path)
    #     image = ImageUtils.loadImage(image_path)
    #     image_name = os.path.basename(os.path.splitext(image_path)[0])

    #     # Get image ious with closest bbox
    #     bboxes, polys, heatmap = model.detection(image)
    #     image_ious[image_name] = {}
    #     for keys, values in ground_truths[image_name].items():
    #         if keys == "size":
    #             continue
    #         for value in values:
    #             # nearest_polygon = Metrics.nearest_polygon(value, bboxes)
    #             if keys not in image_ious.keys():
    #                 image_ious[keys] = [Metrics.polygon_IOU(value, bboxes)]
    #             else:
    #                 image_ious[keys].append(Metrics.polygon_IOU(value, bboxes))
    #     print(image_ious)

    all_results = []
    for k, image_path in enumerate(os.listdir(test_path)): # TODO: Can be changed with imutils paths.listimages(path)
        image_path = os.path.join(test_path, image_path)
        image = ImageUtils.loadImage(image_path)

        # Get the outcome of the detection model
        bboxes, polys, heatmap = model.detection(image)

        # crop images according to bboxes
        cropped_images = ImageUtils.crop_image(image[:,:,::-1], bboxes)

        image_results_path = os.path.join(results_path, os.path.basename(image_path).split("-")[0])
        if not os.path.exists(image_results_path):
            os.makedirs(image_results_path, exist_ok = True)

        # Get tesseract results and write cropped images to a folder
        tesseract_results = []
        for e, cropped_image in enumerate(cropped_images):
            cv2.imwrite(os.path.join(image_results_path, f"{e}-det-"+ os.path.basename(image_path)), cropped_image)
            tesseract_results.append(Tesseract.predict(cropped_image))

        # Find the characters from the folders that cropped images were written
        craft_rec_results = model.recognition(image_folder = image_results_path)
        print(tesseract_results)
        print(craft_rec_results)
        filename, file_ext = os.path.splitext(os.path.basename(image_path))
        mask_file = results_path + "/res_" + filename + '_mask.jpg'
        cv2.imwrite(mask_file, heatmap)

        #FileHandler.saveResult(image_path, image[:,:,::-1], bboxes)
        alligned_results = model.allign_char_results(tesseract_results, craft_rec_results)
        all_results.append(alligned_results)



"""
Copyright (c) 2019-present NAVER Corp.

Model taken from https://github.com/clovaai/deep-text-recognition-benchmark
"""

import torch.nn as nn

from ..utils import TPS_SpatialTransformerNetwork
from ..backbones import RCNN_FeatureExtractor, ResNet_FeatureExtractor
from .blocks import BidirectionalLSTM, Attention


class ReNBAtt(nn.Module):

    __backbones__ = {
                        "rcnn": RCNN_FeatureExtractor,
                        "resnet": ResNet_FeatureExtractor
                    }

    def __init__(self, num_class:int, hidden_size:int = 256, imgH:int = 32, imgW:int = 100, num_fiducial:int = 20,
                    input_channel:int = 1, output_channel:int = 512, backbone:str = "resnet", batch_max_len:int = 25):
        super(ReNBAtt, self).__init__()
        self.num_class = num_class
        self.hidden_size = hidden_size
        self.imgH, self.imgW = imgH, imgW
        self.num_fiducial = num_fiducial
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.backbone = backbone
        self.batch_max_len = batch_max_len


        """ Transformation """
        self.Transformation = TPS_SpatialTransformerNetwork(
                F=self.num_fiducial, I_size=(imgH, imgW), I_r_size=(imgH, imgW), I_channel_num=input_channel)

        """ FeatureExtraction """
        assert self.backbone in ReNBAtt.__backbones__.keys(), ("Given backbone is not in the backbone list"
                                                            f"Supported backbones : {list(ReNBAtt.__backbones__.keys())}")
        self.FeatureExtraction = ReNBAtt.__backbones__[self.backbone](input_channel, output_channel)

        self.FeatureExtraction_output = self.output_channel  # int(imgH/16-1) * 512
        self.AdaptiveAvgPool = nn.AdaptiveAvgPool2d((None, 1))  # Transform final (imgH/16-1) -> 1

        """ Sequence modeling"""
        self.SequenceModeling = nn.Sequential(
            BidirectionalLSTM(self.FeatureExtraction_output, self.hidden_size, self.hidden_size),
            BidirectionalLSTM(self.hidden_size, self.hidden_size, self.hidden_size))
        self.SequenceModeling_output = self.hidden_size

        """ Prediction """
        self.Prediction = Attention(self.SequenceModeling_output, self.hidden_size, self.num_class)

    def forward(self, input, text, is_train=True):
        """ Transformation stage """
        input = self.Transformation(input)

        """ Feature extraction stage """
        visual_feature = self.FeatureExtraction(input)
        visual_feature = self.AdaptiveAvgPool(visual_feature.permute(0, 3, 1, 2))  # [b, c, h, w] -> [b, w, c, h]
        visual_feature = visual_feature.squeeze(3)

        """ Sequence modeling stage """
        contextual_feature = self.SequenceModeling(visual_feature)


        """ Prediction stage """
        # if self.stages['Pred'] == 'CTC':
        #     prediction = self.Prediction(contextual_feature.contiguous())
        # else:
        prediction = self.Prediction(contextual_feature.contiguous(), text, is_train, batch_max_length = self.batch_max_len)

        return prediction



import numpy as np
from typing import List
from shapely.geometry import Polygon

class Metrics:

    @staticmethod
    def shape_shifter(arr):
        return [(arr[0,0], arr[0,1]), (arr[1,0], arr[0,1]), (arr[1,0], arr[1,1]), (arr[0,0], arr[1,1])]

    @staticmethod
    def polygon_IOU_old(arr1:np.array, arr2:np.array):
        """
        Calculate the intersection over union metric for two polygons.

        Args:
            polygon_1 ([type]): [description]
            polygon_2 ([type]): [description]
        """
         # areas = [Metrics.calc_area(arr) for arr in [arr1, arr2]]
        print(arr1)
        print(arr2)
        if isinstance(arr1, np.ndarray):
            arr1 = Metrics.shape_shifter(arr1)
        if isinstance(arr2, np.ndarray):
            arr2 = Metrics.shape_shifter(arr2)
        poly1 = Polygon(arr1)
        poly2 = Polygon(arr2)

        intersection = poly1.intersection(poly2).area

        return  intersection / (poly1.area + poly2.area - intersection)

    @staticmethod
    def polygon_IOU(target:np.array, bboxes):
        """
        Calculate the intersection over union metric for two polygons.

        Args:
            polygon_1 ([type]): [description]
            polygon_2 ([type]): [description]
        """

        if isinstance(target, np.ndarray):
            target = Metrics.shape_shifter(target)

        poly1 = Polygon(target)

        if isinstance(bboxes, list) and len(bboxes) == 4:
            poly2 = Polygon(bboxes)
            intersection = poly1.intersection(poly2).area
            return intersection / (poly1.area + poly2.area - intersection)

        iou_final = []
        for bbox in bboxes:
            if isinstance(bbox, np.ndarray) and bbox.shape[1] == 2:
                bbox = Metrics.shape_shifter(bbox)

            poly2 = Polygon(bbox)
            intersection = poly1.intersection(poly2).area

            iou = intersection / (poly1.area + poly2.area - intersection)
            iou_final.append(iou)

        return sorted(iou_final, reverse = True)[0]

    @staticmethod
    def nearest_polygon(target_arr:np.array, candidate_arrs:List[np.array]):
        """[summary]

        Args:
            target_polygon (np.array): [description]
            candidate_polygons (List[np.array]): [description]
        """
        if target_arr.shape[1] == 2:
            target_arr = Metrics.shape_shifter(target_arr)

        target_p_dist = Polygon(target_arr).centroid
        dist = 100000

        for cand in candidate_arrs:
            if cand.shape[1] == 2:
                cand = Metrics.shape_shifter(cand)
            cand_p_dist = Polygon(cand).centroid
            cand_distance = np.linalg.norm(target_p_dist - cand_p_dist)
            if cand_distance < dist:
                nearest = cand
                dist = cand_distance
        return nearest


import os
import sys
import re
import six
import math
import torch

from natsort import natsorted
from PIL import Image
import numpy as np
from torch.utils.data import Dataset, ConcatDataset, Subset
from torch._utils import _accumulate
import torchvision.transforms as transforms


class RawDataset(Dataset):

    def __init__(self, root, rgb:bool = False, imgH:int = 32, imgW:int = 100):
        self.rgb = rgb
        self.imgH, self.imgW = imgH, imgW
        self.image_path_list = []
        for dirpath, dirnames, filenames in os.walk(root):
            for name in filenames:
                _, ext = os.path.splitext(name)
                ext = ext.lower()
                if ext == '.jpg' or ext == '.jpeg' or ext == '.png':
                    self.image_path_list.append(os.path.join(dirpath, name))

        self.image_path_list = natsorted(self.image_path_list)
        self.nSamples = len(self.image_path_list)

    def __len__(self):
        return self.nSamples

    def __getitem__(self, index):

        try:
            if self.rgb:
                img = Image.open(self.image_path_list[index]).convert('RGB')  # for color image
            else:
                img = Image.open(self.image_path_list[index]).convert('L')

        except IOError:
            print(f'Corrupted image for {index}')
            # make dummy image and dummy label for corrupted image.
            if self.rgb:
                img = Image.new('RGB', (self.imgW, self.imgH))
            else:
                img = Image.new('L', (self.imgW, self.imgH))

        return (img, self.image_path_list[index])


class ResizeNormalize(object):

    def __init__(self, size, interpolation=Image.BICUBIC):
        self.size = size
        self.interpolation = interpolation
        self.toTensor = transforms.ToTensor()

    def __call__(self, img):
        img = img.resize(self.size, self.interpolation)
        img = self.toTensor(img)
        img.sub_(0.5).div_(0.5)
        return img


class NormalizePAD(object):

    def __init__(self, max_size, PAD_type='right'):
        self.toTensor = transforms.ToTensor()
        self.max_size = max_size
        self.max_width_half = math.floor(max_size[2] / 2)
        self.PAD_type = PAD_type

    def __call__(self, img):
        img = self.toTensor(img)
        img.sub_(0.5).div_(0.5)
        c, h, w = img.size()
        Pad_img = torch.FloatTensor(*self.max_size).fill_(0)
        Pad_img[:, :, :w] = img  # right pad
        if self.max_size[2] != w:  # add border Pad
            Pad_img[:, :, w:] = img[:, :, w - 1].unsqueeze(2).expand(c, h, self.max_size[2] - w)

        return Pad_img


class AlignCollate(object):

    def __init__(self, imgH=32, imgW=100, keep_ratio_with_pad=False):
        self.imgH = imgH
        self.imgW = imgW
        self.keep_ratio_with_pad = keep_ratio_with_pad

    def __call__(self, batch):
        batch = filter(lambda x: x is not None, batch)
        images, labels = zip(*batch)

        if self.keep_ratio_with_pad:  # same concept with 'Rosetta' paper
            resized_max_w = self.imgW
            input_channel = 3 if images[0].mode == 'RGB' else 1
            transform = NormalizePAD((input_channel, self.imgH, resized_max_w))

            resized_images = []
            for image in images:
                w, h = image.size
                ratio = w / float(h)
                if math.ceil(self.imgH * ratio) > self.imgW:
                    resized_w = self.imgW
                else:
                    resized_w = math.ceil(self.imgH * ratio)

                resized_image = image.resize((resized_w, self.imgH), Image.BICUBIC)
                resized_images.append(transform(resized_image))
                # resized_image.save('./image_test/%d_test.jpg' % w)

            image_tensors = torch.cat([t.unsqueeze(0) for t in resized_images], 0)

        else:
            transform = ResizeNormalize((self.imgW, self.imgH))
            image_tensors = [transform(image) for image in images]
            image_tensors = torch.cat([t.unsqueeze(0) for t in image_tensors], 0)

        return image_tensors, labels


def tensor2im(image_tensor, imtype=np.uint8):
    image_numpy = image_tensor.cpu().float().numpy()
    if image_numpy.shape[0] == 1:
        image_numpy = np.tile(image_numpy, (3, 1, 1))
    image_numpy = (np.transpose(image_numpy, (1, 2, 0)) + 1) / 2.0 * 255.0
    return image_numpy.astype(imtype)


def save_image(image_numpy, image_path):
    image_pil = Image.fromarray(image_numpy)
    image_pil.save(image_path)



from torch import nn

class RCNN_FeatureExtractor(nn.Module):
    """ FeatureExtractor of GRCNN (https://papers.nips.cc/paper/6637-gated-recurrent-convolution-neural-network-for-ocr.pdf) """

    def __init__(self, input_channel, output_channel=512):
        super(RCNN_FeatureExtractor, self).__init__()
        self.output_channel = [int(output_channel / 8), int(output_channel / 4),
                               int(output_channel / 2), output_channel]  # [64, 128, 256, 512]
        self.ConvNet = nn.Sequential(
            nn.Conv2d(input_channel, self.output_channel[0], 3, 1, 1), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # 64 x 16 x 50
            GRCL(self.output_channel[0], self.output_channel[0], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, 2),  # 64 x 8 x 25
            GRCL(self.output_channel[0], self.output_channel[1], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, (2, 1), (0, 1)),  # 128 x 4 x 26
            GRCL(self.output_channel[1], self.output_channel[2], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, (2, 1), (0, 1)),  # 256 x 2 x 27
            nn.Conv2d(self.output_channel[2], self.output_channel[3], 2, 1, 0, bias=False),
            nn.BatchNorm2d(self.output_channel[3]), nn.ReLU(True))  # 512 x 1 x 26

    def forward(self, input):
        return self.ConvNet(input)